#ifndef __DEFCONF_H
#include "defconf.h"
#endif
/*
 *  fdio - FD device I/O  on WIN32
 *
 *	Written by tantan
 */

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <winioctl.h>
#include <string.h>
#include <ctype.h>
#include <memory.h>
#include <errno.h>
#include <time.h>
#include "defs.h"
#include "misc.h"

void *xmalloc (unsigned size);


BOOL GetDiskGeometry(HANDLE hDisk, PDISK_GEOMETRY lpGeometry )
{
    DWORD ReturnedByteCount;

    return DeviceIoControl(
                hDisk,
                IOCTL_DISK_GET_DRIVE_GEOMETRY,
                NULL,
                0,
                lpGeometry,
                sizeof(*lpGeometry),
                &ReturnedByteCount,
                NULL
                );
}




BOOL
LockVolume(HANDLE hDisk)
{
    DWORD ReturnedByteCount;

    return DeviceIoControl(
                hDisk,
                FSCTL_LOCK_VOLUME,
                NULL,
                0,
                NULL,
                0,
                &ReturnedByteCount,
                NULL
                );
}

BOOL
UnlockVolume(HANDLE hDisk)
{
    DWORD ReturnedByteCount;

    return DeviceIoControl(
                hDisk,
                FSCTL_UNLOCK_VOLUME,
                NULL,
                0,
                NULL,
                0,
                &ReturnedByteCount,
                NULL
                );
}

BOOL
DismountVolume(HANDLE hDisk)
{
    DWORD ReturnedByteCount;

    return DeviceIoControl(
                hDisk,
                FSCTL_DISMOUNT_VOLUME,
                NULL,
                0,
                NULL,
                0,
                &ReturnedByteCount,
                NULL
                );
}


/*
 * Open tape device
 */

	char Drive[]={'\\','\\','.','\\','a',':',0};
	HANDLE hDrive;
	DISK_GEOMETRY Geometry;
	void *FD_Buffer;
	int BufSize;
	char FD_mode;

/* open and read all data to buffer */
global int open_tape(char *file, char mode)
{
	char *p;
	BOOL b;
	int BytesRead;
	
	p = file;
	if (*p == ':')
		p++;
	else if (strmatch(p, "/dev/rfd"))
		p += 8;
	if (!isdigit(*p))
		goto err;
	Drive[4] = *p++ - '0'+'a';

	FD_mode = mode;
/*
 Open and Lock the drive
*/

	hDrive = CreateFile(Drive,
					GENERIC_READ | GENERIC_WRITE,
					0,
					NULL,
					OPEN_EXISTING,
					0,
					NULL );
/*
	hDrive = CreateFile(Drive,
					GENERIC_READ | GENERIC_WRITE,
					0,
					NULL,
					CREATE_ALWAYS,
					0,
					NULL );
*/
	if ( hDrive == INVALID_HANDLE_VALUE ) {
		fprintf(stderr,"tar: Open %s failed %d\n",Drive+4,GetLastError());
		ExitProcess(1);
	}
	if ( LockVolume(hDrive) == FALSE ) {  /* Lock Volume */
		fprintf(stderr,"tar: LockVolume %s failed %d\n",Drive+4,GetLastError());
		ExitProcess(1);
	}
	if ( !GetDiskGeometry(hDrive,&Geometry) ) {
		fprintf(stderr,"tar: GetDiskGeometry %s failed %d\n",Drive+4,GetLastError());
		ExitProcess(1);
	}
/*
	if ( !GetSupportedGeometrys(hDrive) ) {
		fprintf(stderr,"tar: GetSupportedGeometrys %s failed %d\n",Drive+4,GetLastError());
		ExitProcess(1);
	}
*/

	BufSize = Geometry.Cylinders.LowPart *
				Geometry.TracksPerCylinder *
				Geometry.SectorsPerTrack *
				Geometry.BytesPerSector;
	FD_Buffer = xmalloc(BufSize);
	memset(FD_Buffer,0,BufSize);
	if (mode == 'r'){
		b = ReadFile(hDrive,FD_Buffer, BufSize, &BytesRead, NULL);
		if (!b || !BytesRead){
			printf("MFMT: Fatal Read Error %d\n",GetLastError());
			ExitProcess(1);
		}
		DismountVolume(hDrive);
		UnlockVolume(hDrive);
	}

	return 0;
err:
	errno = EINVAL;
	return (-1);

}


/*
 * Read size bytes into buff from tape-device
 */
static int read_point=0;
global int read_tape(char *buff, int size)
{
	
	if (size+read_point <= BufSize){
		memcpy(buff,(char *)FD_Buffer+read_point,size);
		read_point += size;
		return size;
	}else{
		read_point = 0;
		return 0;
	}
}

/*
 * Write size bytes from buff into tape-device
 */
static int write_point=0;
global int write_tape(char *buff, int size)
{
	
	if (size+write_point <= BufSize){
		memcpy((char *)FD_Buffer+write_point,buff,size);
		write_point += size;
		return size;
	}else{
		write_point = 0;
		return 0;
	}
}


/*
 * Backward n bytes
 */
global int back_tape(int n)
{
	if (FD_mode == 'r')
		read_point -= n;
	else if (FD_mode == 'w')
		write_point -= n;

	return (0);
}

/*
 * Close tape device
 */
global int close_tape()
{
	BOOL b;
	int BytesWritten;

	if (FD_mode == 'w'){
		b = WriteFile(hDrive,FD_Buffer, BufSize, &BytesWritten, NULL);
		if ( !b || ( BufSize != BytesWritten ) ) {
			fprintf(stderr,"tar: Fatal Write Error %d\n",GetLastError());
			ExitProcess(1);
		}
		DismountVolume(hDrive);
		UnlockVolume(hDrive);
		
	}
	free(FD_Buffer);
	CloseHandle(hDrive);
	return (0);
}
